""" A suite of functions to handle fault geometries for ground motion model distance calculations.
"""

import numpy as np
import geopy
from geopy.distance import GeodesicDistance

__author__ = 'A. Renmin Pretell Ductram'

#===================================================================================================
# Get fault URC
#===================================================================================================
def get_fault_URC(ULC_lat,ULC_lon,segm_length,segm_strike):

	"""
	Parameter
	=========
	ULC_lat/ULC_lon: Fault segment upper left corner coordinates in deg.
	segm_length: Fault segment lengths.
	segm_strike: Fault segment strike angles in deg. 

	Returns
	=======
	Fault URC lat/lon.
	"""
	
	try:
		N_segm = len(ULC_lat)
	except TypeError:
		N_segm  = 1
		ULC_lat = np.array([ULC_lat])
		ULC_lon = np.array([ULC_lon])
		segm_length = np.array([segm_length])
		segm_strike = np.array([segm_strike])

	URC_lat = np.zeros(N_segm)
	URC_lon = np.zeros(N_segm)
	
	for i in range(N_segm):
		origin = geopy.Point(ULC_lat[i],ULC_lon[i])
		URC    = GeodesicDistance(kilometers=(segm_length[i])).destination(origin,segm_strike[i])
		URC_lat[i] = URC.latitude
		URC_lon[i] = URC.longitude

	return URC_lat,URC_lon

#===================================================================================================
	# Get fault style
#===================================================================================================
def get_faulting_style(rake):

	"""
	Parameter
	=========
	rake: Fault rake in deg.

	Returns
	=======
	Flags: f_normal, f_reverse, and fault mechanism label
	"""

	if (rake>=-150.)*(rake<-30.):
		f_nor = 1
		f_rev = 0
		f_mec = 'NS'
	elif (rake>=30.)*(rake<150.):
		f_nor = 0
		f_rev = 1
		f_mec = 'RS'
	else:
		f_nor = 0
		f_rev = 0
		f_mec = 'SS'

	return f_nor,f_rev,f_mec

#===================================================================================================
# Get representat fault
#===================================================================================================
def get_representative_fault(segm_rake,segm_width,segm_length,segm_dip,segm_Ztor):

	"""
	Parameter
	=========
	segm_rake: Array of segment rake values.
	segm_width: Array of segment width values.
	segm_length: Array of segment length values.
	segm_dip: Array of segment dip values.
	segm_Ztor: Array of segment depth-to-top-of-rupture values.

	Returns
	=======
	Fault representative rake, width, fip, and Ztor in same units as input. 
	"""
	
	try:
		N_segm = len(segm_rake)
	except TypeError:
		N_segm = 1

	if N_segm == 1:
		return segm_rake, segm_width, segm_dip, segm_Ztor
	else:
		frake = segm_rake[0]
		boo = [rake == frake for rake in segm_rake]
		if np.sum(boo) != 0:
			frake = segm_rake[segm_length==np.max(segm_length)][0]

		fwidth = np.sum(segm_length*segm_width)/np.sum(segm_length)
		fdip   = np.sum(segm_length*segm_dip)/np.sum(segm_length)
		fZtor  = np.sum(segm_length*segm_Ztor)/np.sum(segm_length)

		return frake,fwidth,fdip,fZtor



""" Ground motion model by Campbell and Bozorgnia (2019) for Ia and CAV . 
"""

import numpy as np
cimport numpy as np
cimport cython
import CampbellBozorgnia2014
from libc.math cimport pi, cos, pow, sqrt, exp, fmax, log

__author__ = 'A. Renmin Pretell Ductram'

#===================================================================================================
# Ia (0)/CAV (1): phi
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_phi_cy(int IM,double M,double Vs30,double A_1100):
    
    cdef double k_1, k_2, c, n, phi_lnAF, phi_1, phi_2, rho_1, rho_2
    cdef double phi_lnY, rho_lnPGA_lnY, alpha, phi_lnPGA_B, phi
    
    if IM == 1:
        k_1      =  400.
        k_2      = -1.311
        c        =  1.88
        n        =  1.18
        phi_lnAF =  0.3 
        phi_1    =  0.514
        phi_2    =  0.394
        rho_1    =  0.842
        rho_2    =  0.78
    elif IM == 0:
        k_1      =  400.
        k_2      = -1.982
        c        =  1.88
        n        =  1.18
        phi_lnAF =  0.616
        phi_1    =  1.174
        phi_2    =  0.809
        rho_1    =  0.948
        rho_2    =  0.911

    if M <= 4.5:
        phi_lnY       = phi_1*1
        rho_lnPGA_lnY = rho_1*1
    elif M <= 5.5:
        phi_lnY = phi_2+(phi_1-phi_2)*(5.5-M)
        rho_lnPGA_lnY = rho_2+(rho_1-rho_2)*(5.5-M)
    else:
        phi_lnY       = phi_2*1
        rho_lnPGA_lnY = rho_2*1
    phi_lnY_B = sqrt(pow(phi_lnY,2)-pow(phi_lnAF,2))

    alpha = 0
    if Vs30 < k_1:
        alpha = k_2*A_1100*(pow(A_1100+c*pow(Vs30/k_1,n),-1)-pow(A_1100+c,-1))
    
    phi_lnPGA_B = CampbellBozorgnia2014.get_phi_cy(M)

    return sqrt(pow(phi_lnY_B,2)+pow(phi_lnAF,2)+pow(alpha,2)*pow(phi_lnPGA_B,2)+2*alpha*rho_lnPGA_lnY*phi_lnY_B*phi_lnPGA_B)

#===================================================================================================
# Ia (0)/CAV (1): phi
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_tau_cy(int IM,double M,double Vs30,double A_1100):

    cdef double k_1, k_2, c, n, phi_lnAF, phi_1, phi_2, rho_1, rho_2
    cdef double tau_lnY, rho_lnPGA_lnY, alpha, tau_lnPGA_B, tau

    if IM == 1:
        k_1   =  400.
        k_2   = -1.311
        c     =  1.88
        n     =  1.18
        tau_1 =  0.276
        tau_2 =  0.257
        rho_1 =  0.842
        rho_2 =  0.78
    elif IM == 0:
        k_1   =  400.
        k_2   = -1.982
        c     =  1.88
        n     =  1.18
        tau_1 =  0.614
        tau_2 =  0.435
        rho_1 =  0.948
        rho_2 =  0.911

    if M <= 4.5:
        tau_lnY       = tau_1*1
        rho_lnPGA_lnY = rho_1*1
    elif M <= 5.5:
        tau_lnY       = tau_2+(tau_1-tau_2)*(5.5-M)
        rho_lnPGA_lnY = rho_2+(rho_1-rho_2)*(5.5-M)
    else:
        tau_lnY       = tau_2*1
        rho_lnPGA_lnY = rho_2*1
    tau_lnY_B = tau_lnY*1

    alpha = 0
    if Vs30 < k_1:
        alpha = k_2*A_1100*(pow(A_1100+c*pow(Vs30/k_1,n),-1)-pow(A_1100+c,-1))

    tau_lnPGA_B = CampbellBozorgnia2014.get_tau_cy(M)
        
    return sqrt(pow(tau_lnY_B,2)+pow(alpha,2)*pow(tau_lnPGA_B,2)+2*alpha*rho_lnPGA_lnY*tau_lnY_B*tau_lnPGA_B)

#===================================================================================================
# Ia (0)/CAV (1): mu, phi, tau
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def CampbellBozorgnia2019_cy(int IM,double M,double fwidth,double fdip,double Z_tor, double Z_hyp,
                             double[:] R_JB,double[:] R_rup,double[:] R_x,double[:] Vs30,double[:] Z2p5,
                             int fnm,int frv,region):
	
    """
    Parameter
    =========
    IM: 0 for Ia, and 1 for CAV.
    M: Earthquake magnitude.
    fwidth: Fault width in km.
    fdip: Fault dip in deg.
    Z_hyp: Hypocentral depth in km.
    Z_tor: Depth to top of fault rupture in km.
    R_JB: Joyner-Boore distance in km.
    R_rup: Rupture distance in km.
    R_x: Rx distance in km.
    Vs30: Site Vs30. 
    Z2p5: Depth to Vs = 2.5 km/s.
    fnm: 1 for normal/normal-oblique fault, 0 otherwise.
    frv: 1 for reverse/reverse-oblique fault, 0 otherwise.
    region: Earthquake country.

    Returns
    =======
    mean, phi, tau for the selected IM.
    """

    cdef int N_sites = len(R_JB)
    cdef double Vs30_R = 1100.
    cdef double Z2p5_R = exp(7.089-1.144*log(1100.))
    cdef double[:] mu_im  = np.zeros(N_sites, dtype='float64')
    cdef double[:] phi_im = np.zeros(N_sites, dtype='float64')
    cdef double[:] tau_im = np.zeros(N_sites, dtype='float64')
    cdef double mu_pga
    cdef int i
    
    for i in range(N_sites):
        mu_pga = CampbellBozorgnia2014.get_mu_cy(2,M,fwidth,fdip,Z_tor,Z_hyp,R_JB[i],R_rup[i],R_x[i],Vs30_R,Z2p5_R,0.,fnm,frv,region)
        mu_im[i]  = CampbellBozorgnia2014.get_mu_cy(IM,M,fwidth,fdip,Z_tor,Z_hyp,R_JB[i],R_rup[i],R_x[i],Vs30[i],Z2p5[i],exp(mu_pga),fnm,frv,region)
        phi_im[i] = get_phi_cy(IM,M,Vs30[i],exp(mu_pga))
        tau_im[i] = get_tau_cy(IM,M,Vs30[i],exp(mu_pga))

    return np.asarray(mu_im),np.asarray(phi_im),np.asarray(tau_im)



""" A suite of functions to compute distances for ground motion model calculations.
"""

import numpy as np
cimport numpy as np
import nvector as nv
cimport cython
from libc.math cimport pi, sin, cos, tan, atan, atan2, pow, sqrt, isinf

__author__ = 'A. Renmin Pretell Ductram'

#===================================================================================================
# Get min from an array of numbers
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def argmin_cy(double[:] array_A):

	cdef int N = len(array_A)
	cdef double min_val   = array_A[0]
	cdef int min_index = 0

	for i in range(N):
		if array_A[i] < min_val:
			min_val = array_A[i]
			min_index = i
	return min_index

#===================================================================================================
# Get min from an array of numbers
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef min_cy(double[:] values):
	cdef int Nvals = len(values)
	cdef int i
	cdef double min_val
	min_val = 1e9
	for i in range(Nvals):
		if min_val > values[i]:
			min_val = values[i]
	return min_val

#===================================================================================================
# Distance between two locations on the earth
#===================================================================================================
# Cythonized version of the vicenty function by Maurycy Pietrzak: https://github.com/maurycyp/vincenty
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef vincenty_cy(double lat1, double lon1, double lat2, double lon2):

	"""
	Parameter
	=========
	lat/lon: Site coordinates in deg.
	
	Returns
	=======
	Distance in km.
	"""
	
	cdef int    a = 6378137  # meters
	cdef double f = 1 / 298.257223563
	cdef double b = 6356752.314245
	cdef int max_iter = 200
	cdef double conv  = 1e-12
	cdef double sinLambda, cosLambda, sinSigma, cosSigma, sigma, sinAlpha, cosSqAlpha, C, LambdaPrev

	if (lat1-lat2) == 0 and (lon1-lon2) == 0:
		return 0.0

	lat1 = lat1*pi/180
	lat2 = lat2*pi/180
	lon1 = lon1*pi/180
	lon2 = lon2*pi/180

	cdef double U1     = atan((1-f)*tan(lat1))
	cdef double U2     = atan((1-f)*tan(lat2))
	cdef double Lambda = lon2-lon1
	cdef double L      = lon2-lon1
	cdef double sinU1  = sin(U1)
	cdef double cosU1  = cos(U1)
	cdef double sinU2  = sin(U2)
	cdef double cosU2  = cos(U2)

	for iteration in range(max_iter):
		sinLambda = sin(Lambda)
		cosLambda = cos(Lambda)
		sinSigma  = sqrt(pow(cosU2*sinLambda,2)+pow(cosU1*sinU2-sinU1*cosU2*cosLambda,2))
		
		if sinSigma == 0:
			return 0.0
		
		cosSigma   = sinU1*sinU2+cosU1*cosU2*cosLambda
		sigma      = atan2(sinSigma,cosSigma)
		sinAlpha   = cosU1*cosU2*sinLambda/sinSigma
		cosSqAlpha = 1-pow(sinAlpha,2)
		
		try:
			cos2SigmaM = cosSigma-2*sinU1*sinU2/cosSqAlpha
		except ZeroDivisionError:
			cos2SigmaM = 0
		C = f/16*cosSqAlpha*(4+f*(4-3*cosSqAlpha))
		LambdaPrev = Lambda
		Lambda     = L+(1-C)*f*sinAlpha*(sigma+C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*pow(cos2SigmaM,2))))
		
		if abs(Lambda-LambdaPrev)<conv:
			break
	else:
		return None

	cdef double uSq    = cosSqAlpha*(pow(a,2)-pow(b,2))/pow(b,2)
	cdef double A      = 1+uSq/16384*(4096+uSq*(-768+uSq*(320-175*uSq)))
	cdef double B      = uSq/1024*(256+uSq*(-128+uSq*(74-47*uSq)))
	cdef double dSigma = B*sinSigma*(cos2SigmaM+B/4*(cosSigma*(-1+2*pow(cos2SigmaM,2))-B/6*cos2SigmaM*(-3+4*pow(sinSigma,2))*(-3+4*pow(cos2SigmaM,2))))
	cdef double s      = b*A*(sigma-dSigma)/1000

	return s

#===================================================================================================
# Supporting functions
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_bearing_cy(double lat1, double lon1, double lat2, double lon2):
	cdef conv = pi/180
	lat1 = lat1*conv
	lon1 = lon1*conv
	lat2 = lat2*conv
	lon2 = lon2*conv
	cdef double y = sin(lon2-lon1)*cos(lat2) 
	cdef double x = cos(lat1)*sin(lat2)-sin(lat1)*cos(lat2)*cos(lon2-lon1)
	cdef double theta   = atan2(y,x)
	cdef double bearing = (theta+2*pi)%(2*pi) #RP - Edit that might lead to errors
	return bearing

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef cll2xy(double flon, double flat, double[:] slon, double[:] slat):

	cdef double lat, lon, distance, bearing
	cdef int i       = 0
	cdef int N_sites = len(slon)
	cdef double[:] X = np.zeros(N_sites, dtype='float64')
	cdef double[:] Y = np.zeros(N_sites, dtype='float64')

	for lat,lon in zip(slat,slon):
		distance = vincenty_cy(flat,flon,lat,lon)
		bearing  = get_bearing_cy(flat,flon,lat,lon)
		X[i] = distance*sin(bearing)
		Y[i] = distance*cos(bearing)
		i += 1
	return X,Y

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef d2t(P):

	cdef double a1, b1, c1, d1, e1, a2, b2, c2, d2, e2, detinv, s1, s2, t1, t2, f, dclst, dis1, dis2, dis3, dis4
	cdef int inside = 0
	cdef int iclst

	cdef double[:] B        = np.zeros(3, dtype = 'float64')
	cdef double[:] E0       = np.zeros(3, dtype = 'float64')
	cdef double[:] E1       = np.zeros(3, dtype = 'float64')
	cdef double[:,:] E_clst = np.zeros([3,4], dtype = 'float64')

	P_clst = []

	B[0]  = P[0,0]
	B[1]  = P[1,0]
	B[2]  = P[2,0]
	E0[0] = P[0,1]-B[0]
	E0[1] = P[1,1]-B[1]
	E0[2] = P[2,1]-B[2]
	E1[0] = P[0,3]-B[0]
	E1[1] = P[1,3]-B[1]
	E1[2] = P[2,3]-B[2]
	a1 = E0[0]*E0[0]+E0[1]*E0[1]+E0[2]*E0[2]
	b1 = E0[0]*E1[0]+E0[1]*E1[1]+E0[2]*E1[2]
	c1 = E1[0]*E1[0]+E1[1]*E1[1]+E1[2]*E1[2]
	d1 = E0[0]*B[0]+E0[1]*B[1]+E0[2]*B[2]
	e1 = E1[0]*B[0]+E1[1]*B[1]+E1[2]*B[2]

	detinv = 1/(a1*c1-b1*d1)
	s1     = (b1*e1-c1*d1)*detinv
	t1     = (b1*d1-a1*e1)*detinv

	if isinf(detinv) == 0 and s1 >= 0 and s1 <= 1 and t1 >= 0 and t1 <= 1 and s1+t1 <= 1:
		inside = 1
		P_clst.append(B[0]+s1*E0[0]+t1*E1[0])
		P_clst.append(B[1]+s1*E0[1]+t1*E1[1])
		P_clst.append(B[2]+s1*E0[2]+t1*E1[2])
		dclst = P_clst[0]**2+P_clst[1]**2+P_clst[2]**2
		dclst = (dclst)**0.5
		return dclst, P_clst

	B[0]  = P[0,2]
	B[1]  = P[1,2]
	B[2]  = P[2,2]
	E0[0] = P[0,1]-B[0]
	E0[1] = P[1,1]-B[1]
	E0[2] = P[2,1]-B[2]
	E1[0] = P[0,3]-B[0]
	E1[1] = P[1,3]-B[1]
	E1[2] = P[2,3]-B[2]
	a2 = E0[0]*E0[0]+E0[1]*E0[1]+E0[2]*E0[2]
	b2 = E0[0]*E1[0]+E0[1]*E1[1]+E0[2]*E1[2]
	c2 = E1[0]*E1[0]+E1[1]*E1[1]+E1[2]*E1[2]
	d2 = E0[0]* B[0]+E0[1]* B[1]+E0[2]* B[2]
	e2 = E1[0]* B[0]+E1[1]* B[1]+E1[2]* B[2]

	detinv = 1/(a2*c2-b2*b2)
	s2 = (b2*e2-c2*d2)*detinv
	t2 = (b2*d2-a2*e2)*detinv

	if isinf(detinv) == 0 and s2 >= 0 and s2 <= 1 and t2 >= 0 and t2 <= 1 and s2+t2 <= 1:
		inside = 2
		P_clst.append(B[0]+s2*E0[0]+t2*E1[0])
		P_clst.append(B[1]+s2*E0[1]+t2*E1[1])
		P_clst.append(B[2]+s2*E0[2]+t2*E1[2])
		dclst = P_clst[0]**2+P_clst[1]**2+P_clst[2]**2
		dclst = (dclst)**0.5
		return dclst, P_clst

	if -d1 < 0:
		f = 0
	elif -d1 > a1:
		f = 1
	else:
		f = -d1/a1

	E_clst[0,0] = P[0,0]+f*(P[0,1]-P[0,0])
	E_clst[1,0] = P[1,0]+f*(P[1,1]-P[1,0])
	E_clst[2,0] = P[2,0]+f*(P[2,1]-P[2,0])

	if d2 < 0:
		f = 0
	elif -d2 > a2:
		f = 1
	else:
		f = -d2/a2

	E_clst[0,1] = P[0,2]+f*(P[0,1]-P[0,2])
	E_clst[1,1] = P[1,2]+f*(P[1,1]-P[1,2])
	E_clst[2,1] = P[2,2]+f*(P[2,1]-P[2,2])

	if -e2 < 0:
		f = 0
	elif -e2 > c2:
		f = 1
	else:
		f = -e2/c2

	E_clst[0,2] = P[0,2]+f*(P[0,3]-P[0,2])
	E_clst[1,2] = P[1,2]+f*(P[1,3]-P[1,2])
	E_clst[2,2] = P[2,2]+f*(P[2,3]-P[2,2])

	if -e1 < 0:
		f = 0
	elif -e1 > c1:
		f = 1
	else:
		f = -e1/c1

	E_clst[0,3] = P[0,0]+f*(P[0,3]-P[0,0])
	E_clst[1,3] = P[1,0]+f*(P[1,3]-P[1,0])
	E_clst[2,3] = P[2,0]+f*(P[2,3]-P[2,0])
	dis1 = E_clst[0,0]**2+E_clst[1,0]**2+E_clst[2,0]**2
	dis2 = E_clst[0,1]**2+E_clst[1,1]**2+E_clst[2,1]**2
	dis3 = E_clst[0,2]**2+E_clst[1,2]**2+E_clst[2,2]**2
	dis4 = E_clst[0,3]**2+E_clst[1,3]**2+E_clst[2,3]**2

	dis_seq = [dis1, dis2, dis3, dis4]
	dclst = np.nanmin(dis_seq)
	iclst = np.argmin(dis_seq)
	dclst = (dclst)**0.5
	P_clst.append(E_clst[0,iclst])
	P_clst.append(E_clst[1,iclst])
	P_clst.append(E_clst[2,iclst])

	return dclst,P_clst

#===================================================================================================
# Joyner-Boore distance
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_Rjb_cy(	double[:] slat, double[:] slon,
				double flat1, double flon1, double flat2, double flon2,
				double fwidth, double fdip, double fZtor):
		
	"""
	Parameter
	=========
	slat/slon: Site coordinates in deg.
	flat1/flon1: Fault ULC coordinates in deg.
	flat2/flon2: Fault URC coordinates in deg.
	fwidth: Fault width in km.
	fdip: Fault dip in deg.
	fZtor: Depth to top of fault rupture in km.
	
	Returns
	=======
	Joyner-Boore distance.
	"""

	cdef int N_sites    = len(slat)
	cdef double conv    = pi/180
	cdef double[:] Rrup = np.zeros(N_sites, dtype = 'float64')
	cdef double[:] Rjb  = np.zeros(N_sites, dtype = 'float64')
	cdef double[:,:] pt = np.zeros([3,4], dtype = 'float64')
	cdef double botd    = fZtor + fwidth*sin(fdip*conv)
	cdef double rwh     = fwidth*cos(fdip*conv)
	cdef double fstrike, tmp_fstrike, dX, dY
	cdef int i_sta

	fdipDir = ''
	Sxy1    = cll2xy(flon1,flat1,slon,slat)
	Sxy2    = cll2xy(flon2,flat2,slon,slat)

	for i_sta in range(N_sites):
		pt[0,0] = -Sxy1[0][i_sta]
		pt[1,0] = -Sxy1[1][i_sta]
		pt[2,0] = 0.0
		pt[0,1] = -Sxy2[0][i_sta]
		pt[1,1] = -Sxy2[1][i_sta]
		pt[2,1] = 0.0

		if fdipDir == '':
			dX = pt[0,1]-pt[0,0]
			dY = pt[1,1]-pt[1,0]
			tmp_fstrike = atan(dY/dX)/conv
			if dX == 0 and dY > 0:
				fstrike = 0
			elif dX == 0 and dY < 0:
				fstrike = 180
			elif dX > 0:
				fstrike = 90-tmp_fstrike
			elif dX < 0:
				fstrike = 270-tmp_fstrike

			if fstrike + 90 >= 360:
				fdipDir = fstrike+90-360
			else:
				fdipDir = fstrike+90

		dX = rwh*sin(fdipDir*conv)
		dY = rwh*cos(fdipDir*conv)
		pt[0,2] = pt[0,1]+dX
		pt[1,2] = pt[1,1]+dY
		pt[2,2] = 0.0
		pt[0,3] = pt[0,0]+dX
		pt[1,3] = pt[1,0]+dY
		pt[2,3] = 0.0
		a = d2t(pt)
		Rjb[i_sta] = a[0]

	return np.asarray(Rjb)

#===================================================================================================
# Rupture distance
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_Rrup_cy(	double[:] slat, double[:] slon,
				double flat1, double flon1, double flat2, double flon2,
				double fwidth, double fdip, double fZtor):
		
	"""
	Parameter
	=========
	slat/slon: Site coordinates in deg.
	flat1/flon1: Fault ULC coordinates in deg.
	flat2/flon2: Fault URC coordinates in deg.
	fwidth: Fault width in km.
	fdip: Fault dip in deg.
	fZtor: Depth to top of fault rupture in km.
	
	Returns
	=======
	Rupture distance.
	"""

	cdef int N_sites    = len(slat)
	cdef double conv    = pi/180
	cdef double[:] Rrup = np.zeros(N_sites, dtype = 'float64')
	cdef double[:] Rjb  = np.zeros(N_sites, dtype = 'float64')
	cdef double[:,:] pt = np.zeros([3,4], dtype = 'float64')
	cdef double botd    = fZtor + fwidth*sin(fdip*conv)
	cdef double rwh     = fwidth*cos(fdip*conv)
	cdef double fstrike, tmp_fstrike, dX, dY
	cdef int i_sta
	
	fdipDir = ''
	Sxy1    = cll2xy(flon1,flat1,slon,slat)
	Sxy2    = cll2xy(flon2,flat2,slon,slat)

	for i_sta in range(N_sites):
		pt[0,0] = -Sxy1[0][i_sta]
		pt[1,0] = -Sxy1[1][i_sta]
		pt[2,0] = -fZtor
		pt[0,1] = -Sxy2[0][i_sta]
		pt[1,1] = -Sxy2[1][i_sta]
		pt[2,1] = -fZtor

		if fdipDir == '':
			dX = pt[0,1]-pt[0,0]
			dY = pt[1,1]-pt[1,0]
			tmp_fstrike = atan(dY/dX)/conv
			if dX == 0 and dY > 0:
				fstrike = 0
			elif dX == 0 and dY < 0:
				fstrike = 180
			elif dX > 0:
				fstrike = 90-tmp_fstrike
			elif dX < 0:
				fstrike = 270-tmp_fstrike
			if fstrike + 90 >= 360:
				fdipDir = fstrike+90-360
			else:
				fdipDir = fstrike+90

		dX = rwh*sin(fdipDir*conv)
		dY = rwh*cos(fdipDir*conv)
		pt[0,2] = pt[0,1]+dX
		pt[1,2] = pt[1,1]+dY
		pt[2,2] = -botd
		pt[0,3] = pt[0,0]+dX
		pt[1,3] = pt[1,0]+dY
		pt[2,3] = -botd
		a = d2t(pt)
		Rrup[i_sta] = a[0]

	return np.asarray(Rrup)

#===================================================================================================
# Rx
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_Rx_cy(	double[:] slat, double[:] slon, double flat1, double flon1, double flat2, double flon2):

	"""
	Parameter
	=========
	slat/slon: Site coordinates in deg.
	flat1/flon1: Fault ULC coordinates in deg.
	flat2/flon2: Fault URC coordinates in deg.
	
	Returns
	=======
	Rx distance.
	"""
	
	cdef int N_sites = len(slon)
	cdef double[:] Rx = np.zeros(N_sites, dtype='float64')
	cdef int i
	
	for i in range(N_sites):
		frame   = nv.FrameE(a=6371008.8, f=0)
		pointA1 = frame.GeoPoint(flat1,flon1,degrees=True)
		pointA2 = frame.GeoPoint(flat2,flon2,degrees=True)
		pointB  = frame.GeoPoint(slat[i],slon[i],degrees=True)
		pathA   = nv.GeoPath(pointA1, pointA2)
		Rx[i]   = pathA.cross_track_distance(pointB, method='greatcircle')/1000
	
	return np.asarray(Rx)

#===================================================================================================
# Get all distances (convenient to handle multi-segment fault)
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_distances_cy(	double[:] site_lat,double[:] site_lon,double[:] ULC_lat,double[:] ULC_lon,double[:] URC_lat,double[:] URC_lon,
						double[:] segm_width,double[:] segm_length,double[:] segm_dip,double[:] segm_strike,double[:] segm_Ztor):
	
	"""
	Parameter
	=========
	site_lat/site_lon: Site coordinates in deg.
	ULC_lat/ULC_lon: Fault ULC coordinates in deg.
	URC_lat/URC_lon: Fault URC coordinates in deg.
	segm_width: Fault segment width in km. Array if multi-segment fault.
	segm_length: Fault segment length in km. Array if multi-segment fault.
	segm_dip: Fault segment dip in deg. Array if multi-segment fault.
	segm_strike: Fault segment strike in deg. Array if multi-segment fault.
	segm_Ztor: Fault segment depth to top of rupture ni km. Array if multi-segment fault.
	
	Returns
	=======
	Joyner-Boore, rupture, and Rx distances.
	"""
	
	cdef int N_segms = len(segm_dip)
	cdef int N_sites = len(site_lat)
	cdef double[:,:] tmp_Rjb  = np.zeros((N_sites,N_segms), dtype='float64')
	cdef double[:,:] tmp_Rrup = np.zeros((N_sites,N_segms), dtype='float64')
	cdef double[:,:] tmp_Rx   = np.zeros((N_sites,N_segms), dtype='float64')
	cdef double[:] ALL_Rjb    = np.zeros(N_sites, dtype='float64')
	cdef double[:] ALL_Rrup   = np.zeros(N_sites, dtype='float64')
	cdef double[:] ALL_Rx     = np.zeros(N_sites, dtype='float64')
	cdef double[:] slat, slon
	cdef int i 

	for i in range(N_segms):
		for j in range(N_sites):
			slat = np.array([site_lat[j]])
			slon = np.array([site_lon[j]])
			tmp_Rjb[j][i]  = get_Rjb_cy(slat,slon,ULC_lat[i],ULC_lon[i],URC_lat[i],URC_lon[i],segm_width[i],segm_dip[i],segm_strike[i])[0]
			tmp_Rrup[j][i] = get_Rrup_cy(slat,slon,ULC_lat[i],ULC_lon[i],URC_lat[i],URC_lon[i],segm_width[i],segm_dip[i],segm_Ztor[i])[0]
			tmp_Rx[j][i]   = get_Rx_cy(slat,slon,ULC_lat[i],ULC_lon[i],URC_lat[i],URC_lon[i])[0]

	for j in range(N_sites):
		ALL_Rjb[j]  = min(tmp_Rjb[j][:])
		ALL_Rrup[j] = min(tmp_Rrup[j][:])
		ALL_Rx[j]   = tmp_Rx[j][argmin_cy(tmp_Rjb[j][:])]

	return np.asarray(ALL_Rjb),np.asarray(ALL_Rrup),np.asarray(ALL_Rx)



""" Ground motion model by Campbell and Bozorgnia (2011) for CAVdp. 
"""

import numpy as np
cimport numpy as np
cimport cython
import CampbellBozorgnia2010
from libc.math cimport pow, sqrt, log, exp, fmin, fmax

__author__ = 'A. Renmin Pretell Ductram'

#===================================================================================================
# CAVdp: mu
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_mu_cy(int PSVcheck,double M,double fdip,double fZtor,double R_JB,double R_rup,double R_x,double Vs30,double Z1p0,double Z2p5,int fnm,int frv):
    
    cdef double lnCAVgm, mu
    cdef double c_0, c_1, c_2, c_3

    if PSVcheck == 1:
        c_0 =  0.0072
        c_1 =  1.115
        c_2 = -0.067
        c_3 = -0.00330
    else:
        c_0 =  0.0152
        c_1 =  1.105
        c_2 = -0.044
        c_3 = -0.00369
    
    lnCAVgm = CampbellBozorgnia2010.get_mu_cy(1,M,fdip,fZtor,R_JB,R_rup,R_x,Vs30,Z2p5,fnm,frv)
    return c_0+c_1*lnCAVgm+c_2*(M-6.5)*(M>=6.5)+c_3*R_rup

#===================================================================================================
# CAVdp: phi
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_phi_cy(int CAVgmType,int PSVcheck,double alpha):

    cdef double c_1 
    cdef double phi_lnCAVdp, phi_lnAF_PGA, phi_lnPGA, phi_lnAF_CAVgm, phi_lnCAVgm, rho_lnCAVgm, phi
    
    if CAVgmType == 0:
        if PSVcheck == 1:
            return 0.147
        else:
            return 0.131
    
    elif CAVgmType == 1:
        if PSVcheck == 1:
            phi_lnCAVdp = 0.439
        else:
            phi_lnCAVdp = 0.430
            
        c_1            = 0 
        phi_lnAF_PGA   = 0.3000
        phi_lnPGA      = 0.478
        phi_lnAF_CAVgm = 0.300
        phi_lnCAVgm    = 0.371
        rho_lnCAVgm    = 0.735
        
        phi_lnPGA_B   = sqrt(pow(phi_lnPGA,2)-pow(phi_lnAF_PGA,2))
        phi_lnCAVgm_B = sqrt(pow(phi_lnCAVgm,2)-pow(phi_lnAF_CAVgm,2))
        
        phi_1 = pow(phi_lnCAVgm_B,2)
        phi_2 = pow(phi_lnAF_CAVgm,2)
        phi_3 = pow(alpha,2)*pow(phi_lnPGA_B,2)
        phi_4 = 2*alpha*rho_lnCAVgm*phi_lnCAVgm_B*phi_lnPGA_B
        
        return sqrt(pow(phi_lnCAVdp,2)+(pow(c_1,2))*(phi_1+phi_2+phi_3+phi_4))

#===================================================================================================
# CAVdp: tau
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef get_tau_cy(int CAVgmType,int PSVcheck):
    
    cdef double c_1
    cdef tau_lnCAVdp, tau_lnCAVgm, tau

    if CAVgmType == 0:
        return 0.115
    
    elif CAVgmType == 1:

        if PSVcheck == 1:
            tau_lnCAVdp = 0.247
        else:
            tau_lnCAVdp = 0.245
        
        c_1         = 0 
        tau_lnCAVgm = 0.196

        return sqrt(pow(tau_lnCAVdp,2)+(pow(c_1,2))*(pow(tau_lnCAVgm,2)))

#===================================================================================================
# CAVdp: mu, phi, tau
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def CampbellBozorgnia2011_cy(double M,double fdip,double fZtor,
                             double[:] R_JB,double[:] R_rup,double[:] R_x,double[:] Vs30,double[:] Z1p0,double[:] Z2p5,
                             int fnm,int frv,int CAVgmType,int PSVcheck):

    """
    Parameter
    =========
    M: Earthquake magnitude.
    fdip: Fault dip in deg.
    fZtor: Depth to top of fault rupture in km.
    R_JB: Joyner-Boore distance in km.
    R_rup: Rupture distance in km.
    R_x: Rx distance in km.
    Vs30: Site Vs30. 
    Z1p0: Depth to Vs = 1 km/s.
    Z2p5: Depth to Vs = 2.5 km/s.
    fnm: '1' for normal/normal-oblique fault, '0' otherwise.
    frv: '1' for reverse/reverse-oblique fault, '0' otherwise.
    CAVgm_type: 0 for recorded, and 1 for gmm-based.
    PSV_check: 0 for no check, and 1 for yes.

    Returns
    =======
    CAVdp mean, phi, tau.
    """
    
    cdef int N_sites = len(R_JB)
    cdef double[:] A_1100 = np.zeros(N_sites, dtype='float64')
    cdef double[:] alpha = np.zeros(N_sites, dtype='float64')
    cdef double[:] mu_im = np.zeros(N_sites, dtype='float64')
    cdef double[:] phi_im = np.zeros(N_sites, dtype='float64')
    cdef double[:] tau_im = np.zeros(N_sites, dtype='float64')
    cdef int i

    for i in range(N_sites):
        A_1100[i] = CampbellBozorgnia2010.get_mu_cy(0,M,fdip,fZtor,R_JB[i],R_rup[i],R_x[i],1100.,Z2p5[i],fnm,frv)
        A_1100[i] = exp(A_1100[i])
        alpha[i]  = CampbellBozorgnia2010.get_alpha_cy(Vs30[i],A_1100[i])
        
        mu_im[i]  = get_mu_cy(PSVcheck,M,fdip,fZtor,R_JB[i],R_rup[i],R_x[i],Vs30[i],Z1p0[i],Z2p5[i],fnm,frv)
        phi_im[i] = get_phi_cy(CAVgmType,PSVcheck,alpha[i])
        tau_im[i] = get_tau_cy(CAVgmType,PSVcheck)

    return np.asarray(mu_im),np.asarray(phi_im),np.asarray(tau_im)



""" Ground motion model by Campbell and Bozorgnia (2010) for CAVgm; functional form used for CAVdp.
"""

import numpy as np
cimport numpy as np
cimport cython
from libc.math cimport pow, sqrt, log, exp, fmin, fmax

__author__ = 'A. Renmin Pretell Ductram'

#===================================================================================================
# PGA [0]/CAVgm [1]
#===================================================================================================
# Note: PGA implemented for Vs30=1100. only. Need to work on "f_site" term for lower Vs30 
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_mu_cy(int IM, double M,double fdip,double fZtor,double R_JB,double R_rup,double R_x,double Vs30,double Z2p5,int fnm,int frv):
    
    cdef double val
    cdef double f_mag, f_dis, f_flt_Z, f_hng_R, f_hng_M, f_site, f_sed, mu
    cdef double c_0, c_1, c_2, c_3, c_4, c_5, c_6, c_7, c_8, c_9, c_10, c_11, c_12, k_1, k_2, k_3, c, n

    if IM == 0: # PGA
        c_0  = -1.715
        c_1  =  0.500
        c_2  = -0.530
        c_3  = -0.262
        c_4  = -2.118
        c_5  =  0.170
        c_6  =  5.600
        c_7  =  0.280
        c_8  = -0.120
        c_9  =  0.490
        c_10 =  1.058
        c_11 =  0.040
        c_12 =  0.610
        k_1  =  865.0
        k_2  = -1.186
        k_3  =  1.839
        c    =  1.880
        n    =  1.180
    elif IM == 1: # CAVgm
        c_0  = -4.354
        c_1  =  0.942
        c_2  = -0.178
        c_3  = -0.346
        c_4  = -1.309
        c_5  =  0.087
        c_6  =  7.240
        c_7  =  0.111
        c_8  = -0.108
        c_9  =  0.362
        c_10 =  2.549
        c_11 =  0.090
        c_12 =  1.277
        k_1  =  400.0
        k_2  = -2.690
        k_3  =  1.000
        c    =  1.880
        n    =  1.180

    # Magnitude term
    if M <= 5.5:
        f_mag = c_0+c_1*M
    elif M <= 6.5:
        f_mag = c_0+c_1*M+c_2*(M-5.5)
    else:
        f_mag = c_0+c_1*M+c_2*(M-5.5)+c_3*(M-6.5)

    # Geometric attenuation term
    f_dis = (c_4+c_5*M)*log(sqrt(pow(R_rup,2)+pow(c_6,2)))

    # Style of faulting term
    f_flt_Z = fmin(fZtor,1)
    f_flt   = c_7*frv*f_flt_Z+c_8*fnm
    
    # Hanging wall term
    f_hng_R = 1
    if R_JB > 0:
        val = R_rup*1
        if fZtor < 1:
            val = fmax(R_rup,sqrt(pow(R_JB,2)+1))
        f_hng_R = (val-R_JB)/val

    f_hng_M = 1
    if M <= 6.:
        f_hng_M = 0
    elif M <= 6.5:
        f_hng_M = 2*(M-6.)
        
    if fZtor >= 20:
        f_hng_Z = 0
    else:
        f_hng_Z = (20-fZtor)/20

    if fdip <= 70:
        f_hng_dip = 1
    else:
        f_hng_dip = (90-fdip)/20
    
    f_hng = c_9*f_hng_R*f_hng_M*f_hng_Z*f_hng_dip

    # Shallow site response term
    if IM == 0: # PGA 
        f_site = (c_10+k_2*n)*log(Vs30/k_1)

    elif IM == 1: # CAVgm
        A_1100 = get_mu_cy(0,M,fdip,fZtor,R_JB,R_rup,R_x,1100.,Z2p5,fnm,frv)
        A_1100 = exp(A_1100)

        if Vs30 < k_1:
            f_site = c_10*log(Vs30/k_1)+k_2*(log(A_1100+c*pow(Vs30/k_1,n))-log(A_1100+c))
        elif Vs30 < 1100.:
            f_site = (c_10+k_2*n)*log(Vs30/k_1)
        else:
            f_site = (c_10+k_2*n)*log(1100./k_1)

    # Basin site response term
    if Z2p5 <= 1:
        f_sed = c_11*(Z2p5-1)
    elif Z2p5 <= 3:
        f_sed = 0
    else:
        f_sed = c_12*k_3*exp(-0.75)*(1-exp(-0.25*(Z2p5-3)))

    # Compute mean
    return f_mag+f_dis+f_flt+f_hng+f_site+f_sed

#===================================================================================================
# Get CB10 alpha
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def get_alpha_cy(double Vs30,double A_1100):
    
    cdef double k_1 =  400.0
    cdef double k_2 = -2.690
    cdef double c   =  1.880
    cdef double n   =  1.180

    if Vs30 < k_1:
        return k_2*A_1100*(pow(A_1100+c*pow(Vs30/k_1,n),-1)-pow(A_1100+c,-1))
    else:
        return 0

#===================================================================================================
# CAVgm: phi
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
def get_phi_cy(double Vs30,double alpha):

    cdef double phi_lnAF_PGA   = 0.3
    cdef double phi_lnPGA      = 0.478
    cdef double phi_lnAF_CAVgm = 0.300
    cdef double phi_lnCAVgm    = 0.371
    cdef double rho_lnCAVgm    = 0.735

    cdef double phi_lnCAVgm_B = sqrt(pow(phi_lnCAVgm,2)-pow(phi_lnAF_CAVgm,2))
    cdef double phi_lnPGA_B   = sqrt(pow(phi_lnPGA,2)-pow(phi_lnAF_PGA,2))

    return sqrt(pow(phi_lnCAVgm_B,2)+pow(phi_lnAF_CAVgm,2)+pow(alpha,2)*pow(phi_lnPGA_B,2)+2*alpha*rho_lnCAVgm*phi_lnCAVgm_B*phi_lnPGA_B)

#===================================================================================================
# CAVgm: tau
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
def get_tau_cy():
    return 0.196

#===================================================================================================
# CAVgm: mu, phi, tau
#===================================================================================================
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def CampbellBozorgnia2010_cy(double M,double fdip,double fZtor,
                             double[:] R_JB,double[:] R_rup,double[:] R_x,double[:] Vs30,double[:] Z2p5,
                             int fnm,int frv):
    
    """
    Parameter
    =========
    M: Earthquake magnitude.
    fdip: Fault dip in deg.
    fZtor: Depth to top of fault rupture in km.
    R_JB: Joyner-Boore distance in km.
    R_rup: Rupture distance in km.
    R_x: Rx distance in km.
    Vs30: Site Vs30. 
    Z2p5: Depth to Vs = 2.5 km/s.
    fnm: '1' for normal/normal-oblique fault, '0' otherwise.
    frv: '1' for reverse/reverse-oblique fault, '0' otherwise.

    Returns
    =======
    CAVgm mean, phi, tau.
    """
    
    cdef int N_sites = len(R_JB)
    cdef double[:] A_1100 = np.zeros(N_sites, dtype='float64')
    cdef double[:] alpha = np.zeros(N_sites, dtype='float64')
    cdef double[:] mu_im = np.zeros(N_sites, dtype='float64')
    cdef double[:] phi_im = np.zeros(N_sites, dtype='float64')
    cdef double[:] tau_im = np.zeros(N_sites, dtype='float64')
    cdef int i

    for i in range(N_sites):
        A_1100[i] = get_mu_cy(0,M,fdip,fZtor,R_JB[i],R_rup[i],R_x[i],1100.,Z2p5[i],fnm,frv)
        A_1100[i] = exp(A_1100[i])
        alpha[i]  = get_alpha_cy(Vs30[i],A_1100[i])
    
        mu_im[i]  = get_mu_cy(1,M,fdip,fZtor,R_JB[i],R_rup[i],R_x[i],Vs30[i],Z2p5[i],fnm,frv)
        phi_im[i] = get_phi_cy(Vs30[i],alpha[i])
        tau_im[i] = get_tau_cy()

    return np.asarray(mu_im),np.asarray(phi_im),np.asarray(tau_im)



from .faulttools import get_fault_URC
from .faulttools import get_faulting_style
from .faulttools import get_representative_fault
from .distancetools.distancetools import get_bearing_cy
from .distancetools.distancetools import get_Rjb_cy
from .distancetools.distancetools import get_Rrup_cy
from .distancetools.distancetools import get_Rx_cy
from .distancetools.distancetools import get_distances_cy
from .gmms import CampbellBozorgnia2010
from .gmms import CampbellBozorgnia2011
from .gmms import CampbellBozorgnia2014
from .gmms import CampbellBozorgnia2019
from .gmms import FoulserPiggottGoda2015



from setuptools import setup, find_packages, Extension

with open("README.md","r", encoding = 'utf-8') as fp:
	readme = fp.read()

setup(
	name="gmms",
	version="0.1.0",
	description="Ground motion models and supporting tools.",
	author="A. Renmin Pretell Ductram",
	author_email='rpretell@unr.edu',
	url="https://github.com/RPretellD/gmms",
    long_description=readme,
    
    packages=find_packages(),
	include_package_data=True,
	
    install_requires=["numpy","Cython"],

	license='MIT',
	keywords='ground motion model',
	classifiers=[
        "Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
	]
)